# required imports
import numpy as np
import xgboost as xgb
import matplotlib.pyplot as plt
import numpy as np
from numpy.linalg import norm
from sentence_transformers import SentenceTransformer
import logging

from . import models

analyzer = models.analyzer

def get_mean_semantice_distance_for_corpus(cor1, cor2, prefix=None):
    """
    Calculates the distance for each pair of documents
    """
    sem_model = SentenceTransformer('sentence-transformers/all-MiniLM-L12-v1')
    cor1_embed = sem_model.encode(cor1, show_progress_bar=False)
    cor2_embed = sem_model.encode(cor2, show_progress_bar=False)

    if len(cor1) != len(cor2):
        logging.error('The two copuses must be in the same size.')
        return
    
    # dist_list = embedded_dist(cor1_embed, cor2_embed)
    # dist_list = cosine(cor1_embed, cor2_embed)
    # print('dist_list:', dist_list.shape)
    # # compute cosine similarity
    cosine_sim = np.sum(cor1_embed * cor2_embed, axis=1)/(norm(cor1_embed, axis=1)*norm(cor2_embed, axis=1))
    cosine_dist = 1 - cosine_sim
    
    # print('cosine_dist:', cosine_dist.shape)
    # dist_list = []
    #
    # for doc1, doc2 in zip(cor1, cor2):
    #     dist_list.append(get_semantice_distance_for_docs(doc1, doc2, sem_model))
    mean_dist = np.mean(cosine_dist)

    # Plotting a histogram
    if prefix:
        plot_hist(data=cosine_dist, xlabel='Semantic Distance', fig_name=f'plots/{prefix}_semantic_hist.pdf')
    
    return mean_dist


def get_semantice_distance_for_docs(doc1, doc2, sem_model):
    """
    Generates sentence embedding and calculates the distance between them
    """
    embed1 = sem_model.encode(doc1, show_progress_bar=False)
    embed2 = sem_model.encode(doc2, show_progress_bar=False)
    dist = embedded_dist(embed1, embed2)
    return dist


def plot_hist(data, xlabel, fig_name):
    """
    Plots a histogram and saves it.
    """
    plt.hist(data, bins=30)
    plt.ylabel('Count')
    plt.xlabel(xlabel)
    plt.savefig(fig_name)


def embedded_dist(embed1, embed2):
    """
    this function recieve 2 embedings and return the cosine dist between the two
    we will use it to show the dist between embedings of a sentence before and after anonymization
    """

    # compute cosine similarity
    cosine = np.dot(embed1, embed2) / (norm(embed1) * norm(embed2))
    embed_dist = 1 - cosine
    return embed_dist


def plot_sentiment_scatter(x, y):
    """
    Plots a scatter plot with colors. 
    blue dot - same sentiment
    red dot - different sentiment
    """
    # Same sentiment
    x_t = x[x*y > 0]
    y_t = y[x*y > 0]

    # Different sentiment
    x_f = x[x*y <= 0]
    y_f = y[x*y <= 0]

    plt.scatter(x_t, y_t, color='royalblue', label='Same sentiment')
    plt.scatter(x_f, y_f, color='firebrick', label='Invert sentiment')

    # plt.scatter(df_21['txt_vader_sentiment_pred'], df_21['anon_txt_vader_sentiment_pred'])
    plt.xlabel('Before annoymization')
    plt.ylabel('After annonymization')
    plt.legend()
    plt.title('Sentiment score before and after annonymization')
    plt.show()






import gensim.downloader as api
import nltk
from nltk.sentiment import SentimentIntensityAnalyzer
import logging

# nltk.download('vader_lexicon')  # download necessary data for sentiment analysis
nltk.download('stopwords', quiet=True)
nltk.download('punkt', quiet=True)

# Instantiate sentiment analyzer
analyzer = SentimentIntensityAnalyzer()


def upload_we_model(model_name):
    # From GloVe
    # model_name = 'glove-twitter-25'

    possible_models = list(api.info()['models'].keys())
    if model_name not in possible_models:
        logging.error(f'The model {model_name} is not known to Gensim')
    # print('possible models:', possible_models)

    try:
        model = api.load(model_name)

    except Exception as e:
        logging.error(f'Could not upload the model {model_name}', e)
        model = None
    
    return model

we_model = upload_we_model(model_name='word2vec-google-news-300')


import numpy as np
import pandas as pd
import scipy
import re
import logging

from sklearn.feature_extraction.text import CountVectorizer
from annoy import AnnoyIndex
from k_means_constrained import KMeansConstrained
from sklearn.decomposition import PCA
from sklearn.feature_extraction.text import TfidfVectorizer

from . import nlp_utils

def reduce_dims(vecs, dims=100):
    """
    Reduces dimensions using PCA.
    Returns: the low-dimensional vectors
    """
    # Reduce dimensions only if the dimensions are big
    if dims < min(vecs.shape):
        pca = PCA(n_components=dims)
        vecs = pca.fit_transform(vecs)
    return vecs


def lemmatize_and_remove_stops(corpus, stop_list):
    """
    Alternative version to nlp_utils.clean_corpus
    """
    ccorpus = []
    for doc in corpus:
        cdoc = nlp_utils.cleaning(doc, stop_list)
        ccorpus.append(cdoc)
        
    return ccorpus


def remove_stops(corpus, stop_list):
    """
    remove stop words only
    """
    
    ccorpus = []
    for doc in corpus:
        if doc:
            doc_tokenize = nlp_utils.nlp(doc)

            txt = []
            for token in doc_tokenize:
                word = re.sub('\W+', '', str(token)).lower()

                if word and (word not in stop_list):
                    txt.append(str(token)) # onclude all the marks that are not words 
        
            clean_doc = ' '.join(txt)

        else:
            clean_doc = doc
        ccorpus.append(clean_doc)

    return ccorpus


def get_bow(corpus, stop_list, lemmatize = True):
    """ Vectorizes the corpus using CountVectorizer """
    
    vectorizer = CountVectorizer(ngram_range=(1,1))
    if stop_list is not None:
        if lemmatize:
            cc = lemmatize_and_remove_stops(corpus, stop_list)
        else:
            cc = remove_stops(corpus, stop_list)
    else:
        logging.warning('Creating BoW without stopwords removal')
        cc = corpus
    # print(corpus)
    # print(cc)

    try:
        # Vectorizing
        count_data = vectorizer.fit_transform(cc)
        voc = vectorizer.get_feature_names_out()

    except Exception as e:
        logging.warning(f'Could not create a bow: {e}')
        count_data, voc = None, None
    return count_data, voc


def get_tfidf(corpus): #, stop_list, clean = True, lemmatize = True):
    """ Vectorizes the corpus using CountVectorizer """

    try:
        # Vectorizing
        tfidf_model = TfidfVectorizer()

        # Fit Model
        tfidf_vectors = tfidf_model.fit_transform(corpus)
        # voc = tfidf_model.get_feature_names_out()
        voc = None  # Not needed

    except Exception as e:
        logging.warning('Could not create a Tf-Idf:', e)
        tfidf_model, voc = None, None
    
    return tfidf_vectors, voc


def get_anonym_degree(docs = None, vecs = None, min_k = None, stop_list=None):
    """ If K not given, returns the minimal current k and the corresponding documents.
        If k is given, return the documents with k or less neighbohrs  """
    
    if docs is not None:
        # Lemmatizing the documents
        count_data, _ = get_bow(docs, stop_list=stop_list, lemmatize=False)
    elif vecs is not None:
        count_data = vecs
    else: # No input docs or vecs
        print('You must supply documents or vectors')
        return
    if count_data is not None:
        # Converting to array if is sparse
        if scipy.sparse.issparse(count_data):
            count_data = count_data.toarray()
        # Converting any number larger than 1 into 1
        count_data[count_data > 1] = 1
        # Counting unique values
        uniq_arr, uniq_cnt = np.unique(count_data, axis=0, return_counts=True)

        if not min_k:
            min_k = min(uniq_cnt)
            # All the unique vectors
            un_anon = uniq_arr[uniq_cnt <= min_k]
        else:
            # All the unique vectors
            un_anon = uniq_arr[uniq_cnt < min_k]
            min_k = min(uniq_cnt) # For the return value

        # Getting the unique vectore indeces
        indeces_list = []
        for row in un_anon:
            # Get the similar rows
            similar_vals = np.where((count_data == (row)).all(axis=1))
            indeces_list.append(similar_vals[0].tolist())
    else:  # count_data is None
        min_k, indeces_list = None, []

    return min_k, indeces_list


def get_diff(vecs):
    """
    Finds the differences between arrays of 0,1.
    The input is either a list of lists or a matrix
    """
    # Creating a matrix if needed
    if isinstance(vecs, list):
        mat = np.vstack(vecs)
    else: 
        mat = vecs
    mat[mat > 0] = 1
    #nparray = sparse_mat.toarray()
    xsum = np.sum(mat, axis=0)

    #print('xsum\n', xsum)

    union_x = (xsum > 0).astype('int')
    inter_x = (xsum == mat.shape[0]).astype('int')

    diff = union_x - inter_x
    return diff


def build_annoy_search_tree(n_list):
    """
    Builds annoy search tree
    n_list: A matrix. rows = documents. columns = vocabulary
    """
    # Build an Annoy index with 10 trees. angular = cosine similarity
    annoy_index = AnnoyIndex(n_list.shape[1], metric='angular')
    for i, x in enumerate(n_list):
        annoy_index.add_item(i, x)
    annoy_index.build(10)

    return annoy_index


def get_k_unused_items(item, annoy_tree, used_items, k):
    """
    Returns k unused items.
    """
    # Find the k nearest neighbors to the first sentence
    k_unused_items = []
    new_k = k

    i = 0  # TEMP

    while len(k_unused_items) < k:
        nearest_neighbors = annoy_tree.get_nns_by_vector(item, new_k)
        for nn in nearest_neighbors:
            if (nn not in used_items) and (len(k_unused_items) < k) and (nn not in k_unused_items):
                k_unused_items.append(nn)
        new_k += 1
        i += 1  # TEMP
    return k_unused_items


def find_k_neighbors_using_annoy(docs, k, dim_reduct = 0, stop_list = None):
    """
    1. Create BoW representation
    2. For each document, finds k nearest neighbors
    Returns a list of indexes.
    """
    if isinstance(docs, np.ndarray): # Documents already embedded
        vecs = docs
    
    else: # A list of texts
        # Cleaning the documents
        cdocs = [nlp_utils.lemmatize_doc(doc, stop_list=stop_list) for doc in docs]
        docs = cdocs
        vecs, _ = get_tfidf(docs)
        vecs = vecs.toarray()  # From sparse matrix to array 
        
        # Reduce dimensions
        if dim_reduct:
            vecs = reduce_dims(vecs=vecs, dims=dim_reduct)

    neighbor_list = []
    annoy_tree = build_annoy_search_tree(vecs)

    logging.debug('Finding k neighbors using Annoy...')

    used_indexes = set([])
    num_docs = vecs.shape[0]
    for i in range(num_docs):
        # To prevent redandent
        if i not in used_indexes:
            similar_doc_ind = get_k_unused_items(vecs[i], annoy_tree, used_indexes, k)
            neighbor_list.append(similar_doc_ind)
            for sd in similar_doc_ind:
                if sd in used_indexes:
                    logging.warning(f'The index {sd} was already used.')
                # Adding the index to the used items
                used_indexes.add(sd)  
            # No more k unused indexes
            if  len(used_indexes) > (num_docs - k):
                break

    return neighbor_list


def reorder_documents(doc_list, neighbor_list, num):
    """
    Re-orders documents after anonymization.
    Returns a list of documents in the original order
    """
    # Creating a list with n Nones
    anonym_docs = [None] * num

    for indexes, docs in zip(neighbor_list, doc_list):
        for i, d in zip(indexes, docs):
            anonym_docs[i] = d
    return anonym_docs


def create_neighbors_df_for_comparison(doc_list, neighbor_list):
    """
    Saves the neighbors in a dataframe
    """
    neighbor_list_docs = [None] * len(neighbor_list)
    for i, nl in enumerate(neighbor_list):
        neighbor_list_docs[i] = list(neighbor_list[i])
        for j, doc_idx in enumerate(nl):
            neighbor_list_docs[i][j] = doc_list[doc_idx]
    return pd.DataFrame(neighbor_list_docs)


def force_anonym(docs, neighbor_list, stop_list):
    """ 
    For each group of neighbors, replace different words in *.
    Returns a list of anonymized documents.
    """
    annon_docs = docs.copy()
 
    cdocs = [nlp_utils.lemmatize_doc(doc, stop_list=stop_list) for doc in docs]
    docs = cdocs

    used_indexes = []
    for neighbors in neighbor_list:
        # print('neighbors:', neighbors)
        curr_docs = []
        for n in neighbors:
            # Adding the document to the similar doc list
            curr_docs.append(docs[n])
        anonym_docs = delete_uncommon_words(curr_docs, stop_list)
        i = 0
        for n in neighbors:
            used_indexes += neighbors
            annon_docs[n] = anonym_docs[i]
            i += 1

    # Removing all the documents without partners
    unused_indexes = list(set(range(len(docs))) - set(used_indexes))
    for i in unused_indexes:
        annon_docs[i] = '*'

    return annon_docs


def add_neighbor_list_to_df(df, neighbor_list):
    """
    Adds the neigbors for each document
    """
    df['neigbors'] = None
    for k_neighbors in neighbor_list:
        for n in k_neighbors:
            df.loc[n, 'neigbors'] = str(k_neighbors)
    return df


def delete_uncommon_words(docs, stop_list):
    """
    Deletes the uncommon words from given documents
    """
    # Lemmatizing
    ldocs = [nlp_utils.lemmatize_doc(d, stop_list=stop_list) for d in docs]

    vecs, voc = get_bow(ldocs, stop_list=stop_list)

    if vecs is not None:
        vecs = vecs.toarray()
        diff = get_diff(vecs)

        words_to_delete = voc[diff > 0]
        
        # logging.debug(f'docs: {docs} \nuncommon words: {words_to_delete}')

        temp_docs = []
        for d in ldocs:
            new_d = d
            for word in words_to_delete:
                #new_d = new_d.replace(word, '*')
                new_d = re.sub(rf'\b{word}\b', '*', new_d, flags=re.IGNORECASE)
            temp_docs.append(new_d)
    else:
        # All stop words. Return lemmatized document
        temp_docs = ldocs
    
    # logging.debug(f'After forcing: {temp_docs}')
    return temp_docs


def delete_word(word_dict, word):
    """
    When forcing anonymization, finds the word in the dictionary, and updates the replaced tag.
    """
    if word in word_dict:
        word_dict[word]['replaced'] = '*'
    else:
        for w in word_dict.keys():
            if (word_dict[w]['lemma'] == word) or (word_dict[w]['replaced'] == word):
                word_dict[w]['replaced'] = '*'
                break  # Exit for


# TEMP - also in LLM-utils
def ckmeans_clustering(docs, k, n_jobs = -1, dim_reduct = 0, stop_list = None):
    """
    Runs k-means with constrains.
    More on constrain-k-means: https://towardsdatascience.com/advanced-k-means-controlling-groups-sizes-and-selecting-features-a998df7e6745
    Returns a list of neighbor list
    """
    logging.debug(f'{len(docs)} documents')
    
    if isinstance(docs, np.ndarray): # Documents already embedded
        vecs = docs
    else: # A list of texts
        # Cleaning the documents
        cdocs = [nlp_utils.lemmatize_doc(doc, stop_list=stop_list) for doc in docs]
        docs = cdocs
        vecs, _ = get_tfidf(docs)

        vecs = vecs.toarray()  # From sparse matrix to array (KMeansConstrained does not work on sparse matrix)

    # Reduce dimensions
    if dim_reduct:
        vecs = reduce_dims(vecs=vecs, dims=dim_reduct)

    # logging.info('Got BoW')
    num_clusters = vecs.shape[0] // k
    min_size = k
    max_size = k

    # For example, if k=3 and there are 100 sequences,
    # allow one cluster with k+1
    mod_data =  vecs.shape[0] % k
    if mod_data != 0:
        max_size += mod_data

    # Clustering parameters
    n_init, max_iter = 6, 100
    
    clf = KMeansConstrained(
     n_clusters=num_clusters,
     size_min=min_size,
     size_max=max_size,
     random_state=0,
     n_jobs=n_jobs,
     n_init=n_init,
     max_iter=max_iter
    )
    # Logging
    logging.debug(f'Clustering... Data dimensions: {vecs.shape}, Parameters: n_init {n_init}   max_iter {max_iter}   n_jobs {n_jobs}')

    try:
        clf.fit_predict(vecs)
        logging.debug('Finish clustering. Getting original indexes.')
        pair_list = []
        for i in range(1, num_clusters):
            curr_pair = np.where(clf.labels_ == (i))[0].tolist()
            if curr_pair not in pair_list:
                pair_list.append(tuple(curr_pair))
    except Exception as e:
        logging.error(f'Could not find neighbors using k_means_constrained: {e}')
        logging.info('Finding K neighbors using Annoy')
        # Sending the already cleaned vectors
        pair_list = find_k_neighbors_using_annoy(docs=None, k=k, vecs=vecs, dim_reduct = dim_reduct)
        
    return pair_list



# Imports
import pandas as pd
import numpy as np
import re
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize
import spacy
import re
import bz2
from collections import defaultdict
import operator
import logging

# Defining some global variables
stopword_list = None

short_stopword_list = None
long_stopword_list = None

nlp = spacy.load('en_core_web_sm', disable=['ner', 'parser'])  # disabling Named Entity Recognition for speed
# we_model = models.we_model

def get_list_from_file(file_name, num=None):
    """"
    Reads a file with words (each word on different line) and returns a list of these words.
    """
    try:
        with open(file_name, 'r') as file:
            
            # reading the file
            data = file.read()
            
            # replacing end splitting the text 
            # when newline ('\n') is seen.
            word_list = data.split("\n")

            # Take only part of the words
            if num and (num < len(word_list)):
                word_list = word_list[:num]
    
    except Exception as e:
        logging.error(f'Could not read the file {file_name}: {e}')
        word_list = None
    return word_list


def corpus_stop_words(corpus,num_stop):
    """this function gets as input the text corpus (corpus) - list of texts, and a number (num_stop),
    it finds the most common words in the corpus and returns them as a dataframe/csv"""
    # Step 1: Tokenize the texts - make sure the corpus was pre-processed!! (lower/etc)
    tokenized_corpus = [text.split() for text in corpus]

    # Step 2: Count unique words per document
    unique_words_per_document = [set(text) for text in tokenized_corpus]

    # Step 3: Count word frequencies
    word_frequency = defaultdict(int)
    for words in unique_words_per_document:
        for word in words:
            word_frequency[word] += 1

    # Step 4: Sort and select top words
    most_frequent_words = sorted(word_frequency.items(), key=operator.itemgetter(1), reverse=True)[:num_stop]

    return most_frequent_words # list of most frequent words to set as stop-words


def reading_bz_file(train_file):
    """ Reading the input file and returns a dataframe """
    # Credit: https://www.kaggle.com/code/pierremegret/gensim-word2vec-tutorial

    # Readling the file to list of comments
    train_file = bz2.BZ2File(train_file)
    train_file_lines = train_file.readlines()

    # Converting from raw binary strings to strings that can be parsed
    train_file_lines = [x.decode('utf-8') for x in train_file_lines]

    # Extracting the labels and sentences
    train_labels = [0 if x.split(' ')[0] == '__label__1' else 1 for x in train_file_lines]
    train_sentences = [x.split(' ', 1)[1][:-1].lower() for x in train_file_lines]  # And converting to lower case

    # Converting to df
    df = pd.DataFrame(list(zip(train_sentences, train_labels)), columns=['txt', 'sentiment'])

    # Adding number of words
    df['num_of_words'] = df['txt'].apply(lambda x: len(x.split(' ')))

    return df


def get_voc(corpus):
    """
    Creates a set of unique words in a given corpus
    """
    word_set = set([])

    for doc in corpus:
            doc_words = set(word_tokenize(doc))
            word_set = word_set.union(doc_words)
    return word_set


def get_lemma(word, nlp=nlp):
    """ Returns the lemma of the given word """
    word = nlp(word)
    txt = [token.lemma_ for token in word]
    lemma = txt[0]
    return lemma


def create_word_dict(corpus, stop_list):
    """ Creates a word dictionary. Keys = words. Values = three boolians - stop/lemma/replaced"""
    all_words = get_voc(corpus)

    logging.debug('Got vocabulary')

    all_word_dict = {}

    for word in all_words:
        # Cleaning non AB characters
        word = replace_non_ab_chars(word)

        # Creating the word dictionary
        word_dict = {
            'protected': None,
            'lemma': None,
            'replaced': None}
        if len(word) > 0:
            if word in stop_list:
                word_dict['protected'] = True
            else:
                lemma = get_lemma(word)
                if  lemma != word:
                    word_dict['lemma'] = lemma
            all_word_dict[word] = word_dict
    return all_word_dict


def print_doc(doc, all_word_dict):
    """ Prints document based on the word dictionary """
    # out_str = 'Legend: (protected) [replaced] {lemmatized}\n'
    out_str = ''
    words = doc.split()

    for w in words:
        print_w = w.lower()
        w = replace_non_ab_chars(w)
        if w in all_word_dict:
            if all_word_dict[w]['protected']:
                print_w = '(' + print_w + ')'
            elif all_word_dict[w]['lemma']:
                print_w = '{' + all_word_dict[w]['lemma'] + '}'
            if all_word_dict[w]['replaced']:
                print_w = '[' + all_word_dict[w]['replaced'] +']'
        out_str = f'{out_str}{print_w} '

    return out_str


def replace_non_ab_chars(word):
    word = re.sub("[^A-Za-z']+", '', str(word)).lower()
    return word


def cleaning(doc, stop_list):
    """Lemmatizing and removes stopwords"""
    # Defining the document
    if doc:
        doc = nlp(doc)

        # Lemmatizes and removes stopwords

        # For some reason, sometimes the lemmatization is not consistent - 
        # identical words receive a different lemmatization (for example, acting is sometimes 
        # changed to act and sometimes remains acting). 
        # Apperantly, Spacy's lemmatization depends on the part of speech: 
        # https://stackoverflow.com/a/74176351
        # To overcome this, we added the break_doc parameter

        txt = []
        for token in doc:
            word = re.sub('\W+', '', str(token)).lower()
            if word and (word not in stop_list):
                txt.append(token.lemma_)
    
        clean_doc = ' '.join(txt)
    else:
        clean_doc = doc

    return clean_doc


def lemmatize_doc(doc, stop_list):
    """Lemmatizes document"""
    # Breaking the document to allow consistent lemmatization (see function cleaning())

    doc = nlp(doc)

    # Lemmatizes and removes stopwords
    # doc needs to be a spacy Doc object
    txt = []
    for token in doc:
        if str(token) in stop_list:  # Add stopword as it is
            txt.append(str(token))
        else:  # Lemmatize other words
            txt.append(token.lemma_)
    clean_doc = ' '.join(txt)

    return clean_doc


def clean_corpus(corpus):
    """ Cleans the corpus """
    brief_cleaning = (re.sub("[^A-Za-z']+", ' ', str(row)).lower() for row in corpus)
    corpus_lemmas = [cleaning(doc) for doc in nlp.pipe(brief_cleaning, batch_size=5000)]
    return corpus_lemmas


def get_general_word_from_cluster(word_list, wemodel):
    """ Finds the most similar words usind word embedding"""
    try:
        # For W2V
        known_words = [w for w in word_list if w in wemodel.vocab]
    except:
        # For GloVe
        known_words = [w for w in word_list if w in wemodel.index_to_key]
    word_vecs = []
    for w in known_words:
        word_vecs.append(wemodel[w])
    word_vecs = np.array(word_vecs)
    
    # Finding the centorid
    centroid = np.mean(word_vecs, axis=0)

    # print('known words:', len(known_words), 'word_list', len(word_list))
    if len(known_words) > 0:
        # we_word = we_model.most_similar(known_words, topn=1)[0][0]
        we_word = wemodel.most_similar([centroid], topn=1)[0][0]
    else:
        we_word = None
    return we_word


def replace_words_in_df(df_0, cluster_dict, distance_dict, word_dict_copy, col, wemodel, stop_list=None):
    """ Replaces the words in the dataframe """

    out_str = f'Number of clusters: {len(cluster_dict)}   num protected words: {len(stop_list)}'
    logging.debug(out_str)

    # Working with a copy of the df:
    df_copy = df_0
    # word_dict_copy = word_dict_0.copy()

    df_copy['anon_txt'] = df_copy[col].apply(lambda x: lemmatize_doc(x, stop_list=stop_list))

    logging.debug('Going over clusters')

    max_size = 0

    for key, words in cluster_dict.items():
        if key >= 0:  # Ignoring the -1 label
            #if len(cluster_dict[key]) < 20:  # so it will make sense!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # Getting the general word
            general_word = get_general_word_from_cluster(words, wemodel).lower()

            if stop_list and (general_word not in stop_list):
                stop_list.append(general_word)  # the list of new words
            rep_str = f'distance: {distance_dict[key]} \tcluster size: {len(words)} \treplacing {words} in {general_word}'
            logging.debug(rep_str)

            if len(words) > max_size: 
                max_size = len(words)

            words_to_replace = []
            for word in words:
                if word not in word_dict_copy:  # Lemmatized word
                    word_dict_copy[word] = {'protected': False, 'lemma': word, 'replaced':False}
                    
                if not word_dict_copy[word]['protected']:
                    # Updaing the word dictionary that the words were replaced
                    word_dict_copy[word]['replaced'] = general_word  # Problem: greate in line 2 miss-identified as replaced
                
                # df_copy['anon_txt'] = df_copy['anon_txt'].replace(fr'\b{word}\b', general_word, regex=True)
                
                words_to_replace.append(word)

            # Replacing whole words
            rep_str = create_rep_string(words_to_replace)
            # print('rep_str 1: ', rep_str)
            # rep_str = '\\b|\\b'.join(words_to_replace)
            # print('rep_str 2: ', rep_str)

            df_copy['anon_txt'] = df_copy['anon_txt'].replace(fr'\b{rep_str}\b', general_word, regex=True)
    logging.debug('Creating history')
    logging.info(f'Largest clone size - {max_size}')
    df_copy['anon_txt_history'] = df_copy[col].apply(lambda x: print_doc(x, word_dict_copy))

    if stop_list:
        logging.debug('Stop word list was updated')

    return df_copy, word_dict_copy, stop_list


def create_rep_string(word_list):
    """
    Creates a string in the form: 'word1|word2|word3..'
    """
    out_str = f'\\b{word_list[0]}'
    for w in word_list[1:]:
        out_str = f'{out_str}\\b|\\b{w}'
    out_str = f'{out_str}\\b'
    return out_str


def get_stat(word_dict):
    """
    Returns the number of stop words, lemmatized and replaced words
    """
    prot_count = 0
    lemm_count = 0
    rep_count = 0

    all = 0

    for word in word_dict.values():
        if word['protected']:
            prot_count += 1
        elif word['replaced']:
            rep_count += 1
        elif word['lemma']:
            lemm_count += 1
        all += 1
    return prot_count, lemm_count, rep_count, all


if __name__ == 'main':
    print('yehhh')




from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
from sentence_transformers import SentenceTransformer
import numpy as np
import logging
from . import anonym_utils


def print_example(indexes, origina_docs, new_docs):
    print('Before:')
    for i in indexes:
        print(origina_docs[i])
    print('\nAfter:')
    for i in indexes:
        print(new_docs[i])


def sum_text_0(doc_list, tokenizer, gen_model):
    # define the input sentences
    #input_text = '. '.join(doc_list)
    input_text = ''
    for doc in doc_list:
       input_text = f'{input_text}\n- {doc}. ' 

    # preprocess the input sentences
    input_ids = tokenizer.encode(f'Generate a general form of these sentences: \n{input_text}', return_tensors="pt")

    # generate the summary sentence
    output_ids = gen_model.generate(input_ids=input_ids, max_length=32, num_beams=4, early_stopping=True)
    output = tokenizer.decode(output_ids[0], skip_special_tokens=True)

    return output


def sum_text(doc_list, tokenizer, gen_model):
    # define the input sentences
    #input_text = '. '.join(doc_list)
    input_text = ''
    i = 1
    for doc in doc_list:
       input_text = f'{input_text}{i}: {doc}. ' 
       i += 1
    
    # print('doc_list:', doc_list)
    # preprocess the input sentences
    prompt = prompt_builder(doc_list)
    input_ids = tokenizer.encode(prompt, return_tensors="pt", show_progress_bar=False)

    # generate the summary sentence
    output_ids = gen_model.generate(input_ids=input_ids, max_length=32, num_beams=4, early_stopping=True)
    output = tokenizer.decode(output_ids[0], skip_special_tokens=True)

    return output


def run_anonymization_on_txt(docs, k, n_jobs):
    """ Finding K nearest neighbors and summarize them """
    
    # Defining models
    # google/flan-t5-xl
    # "google/flan-t5-xl"
    tok_model_name = 'google/flan-t5-xl'
    gen_model_name = 'google/flan-t5-xl'
    emb_model_name = 'sentence-transformers/all-MiniLM-L12-v1'

    logging.debug(f'Tokenizer: {tok_model_name}  Genrative model: {gen_model_name}  Embedding model: {emb_model_name}')
    
    # Uploading models
    tokenizer = AutoTokenizer.from_pretrained(tok_model_name)
    gen_model = AutoModelForSeq2SeqLM.from_pretrained(gen_model_name)
    emb_model = SentenceTransformer(emb_model_name)

    annon_docs = docs.copy()
    # Embedding
    logging.info('Embedding the documents...')
    docs_emb = emb_model.encode(docs, show_progress_bar=False)
    logging.debug(f'Embedding dimensions: {docs_emb.shape}, {type(docs_emb)}')
    # neighbor_list = ckmeans_clustering(docs_emb, k)
    neighbor_list = anonym_utils.ckmeans_clustering(docs_emb, k=k, n_jobs=1, dim_reduct=False)
    logging.info(f'Found {len(neighbor_list)} groups of {k} neighbors')
    # print('&&&&&&&&&&&', 3)

    logging.info(f'Generating alternative documents...')
    for n in neighbor_list:
        # print('n:', n)
        curr_docs = []
        for doc_index in n:
            # Adding the document to the similar doc list
            curr_docs.append(docs[doc_index])
        sum_doc = sum_text(curr_docs, tokenizer, gen_model)
        # sum_doc = sum_text_0(curr_docs, tokenizer, gen_model)
        # print('similar_doc_ind', n, '\tSummary:', sum_doc)
        for doc_index in n:
            annon_docs[doc_index] = sum_doc
    # print('&&&&&&&&&&&', 4)
    
    logging.info(f'Generation completed.')

    return annon_docs, neighbor_list


def prompt_builder(docs):
    """
    Generating the prompt for document generalization
    """

    # Removing /n from documens
    new_docs = []
    for d in docs:
        new_docs.append(d.replace('\n',' '))
    
    # Adding new line and '-' between documnets
    sents = '\n-'.join(new_docs)
    # Adding '-' before the first document
    sents = '-' + sents + '\n'

    prompt = f'''Write a general sentence that best represents each of the following sentences and is true to all of them. For example: 

Sentences:
-the alchemist: sure this is an interesting book. unfortunately the copy we received was written in spanish!
-immensee: a beautiful story, but the translation from the original language (german) is unbelievably poor.
Result: 
Good book, problem with the translation.

Sentences:
-low budget: some good action and scenery, but otherwise pretty low budget movie. only recommended for die-hard horror fans.
-horrible movie!: please listen everybody this movie is terrible-don\'t bother ok!it is just dumb not scary!
Result: 
The movie was bad.

Sentences:
-america the beautiful: this book works great for teaching the song. the hardback cover is great for multiple uses.
-enjoyable music: bought this to use for our church\'s hawaiaan-harvest festival for children ages 4-12. great music!
Result: 
Good and useful music.

Sentences:
-invaluable: the authors have provided the quintessential study guide to the canterbury tales. this book is invaluable.
-magical: wonderful characters, stories within stories. this book offers beautiful writing, a deep spirituality and a happy ending!
Result: 
Wonderful book and great writing.

Sentences:
-textbook: book shipped quickly and was in excellent condition as stated. easy transaction would buy again
-mathbook: excellent condition of book. description is true to the condition of the mathbook.shipped quickly excellent seller.
-good deal!: used library book, but still in good condition. book came quickly and was cheap. overall good deal!
Result: 
Good condition book, good delivery and good seller.

Sentences:
-great: wishmaster is a fantastic album, that should grace everyone's music collection. innovating and always refreshing. truly a masterpiece.
-fantastic debut: a great debut record from a band who deserve a lot of attention. buy this record twice.
-great music, great arrangements and performance.: very satisfied with this cd. cem duruoz is a genius.
Result: 
Great music, highly recommended.

Sentences:
{sents}
Result:
'''
    return prompt





# Imports
import numpy as np
from sklearn.cluster import DBSCAN
import matplotlib.pyplot as plt
import logging

def define_eps_cos(wemodel):
    """Defines distance between pairs of words"""
    
    # Collecting distances of good pairs - cosine similarity
    sim_list_best = get_pairs_sim_cos(get_good_pairs(), wemodel)
    # print('mean similarity between good pairs\t', np.median(sim_list_best))

    # Collecting distances of bad pairs - cosine similarity
    sim_list_worst = get_pairs_sim_cos(get_bad_pairs(), wemodel)
    # print('mean similarity between bad pairs\t', np.median(sim_list_worst))
    
    best_dist = 1 - np.median(sim_list_best)
    worst_dist = 1 - np.median(sim_list_worst)

    # The  threshold for cluster max_dist should be in the middle of the two values above = 0.1765
    threshold = (best_dist + worst_dist)/2
    
    return threshold


def define_eps_euc(wemodel):
    """Defines distance between pairs of words"""
    
    # Collecting distances of good pairs - Euclidean distance
    dist_list_best = get_pairs_dist_euc(get_good_pairs(), wemodel)
    # print('mean distance between good pairs\t', np.median(dist_list_best))

    # Collecting distances of bad pairs - Euclidean distance
    dist_list_worst = get_pairs_dist_euc(get_bad_pairs(), wemodel)
    # print('mean distance between bad pairs\t', np.median(dist_list_worst))
    
    best_dist = np.median(dist_list_best)
    worst_dist = np.median(dist_list_worst)

    # The  threshold for cluster max_dist should be in the middle of the two values above = 0.1765
    threshold = (best_dist + worst_dist)/2
    return threshold


def get_good_pairs():
    """Returns pairs of similar words"""
    best_pairs_ls = [
        ['good','great'],
        ['dog','cat'],
        ['green','yellow'],
        ['dad','mom'],
        ['purchase','buy'],
        ['gift','present'],
        ['fast','quick'],
        ['big','huge'],
        ['item','product'],
        ['text','script'],
        ['john','julie'], # Adding names
        ['shirley','matthew'],
        ['joseph','smith'],
        ['michael','sylvia'],
        ['bonnie','henry'],
        ['paul','jr'],
        ['duncan','kelly']]
    return best_pairs_ls


def get_bad_pairs():
    """Returns pairs of unsimilar words"""
    worst_pairs_ls = [
        ['good','trip'],
        ['think','fat'],
        ['sister','white'],
        ['grammar','small'],
        ['boy','buy'],
        ['playstation','old'],
        ['cd','low'],
        ['battery','cold'],
        ['wonderful','check'],
        ['book','rich'],
        ['brother','unless'],
        ['john','ear'],
        ['sixty','stay'],
        ['barbara','wave'],
        ['buy','niece']]
    return worst_pairs_ls


def get_pairs_sim_cos(pair_list, wemodel):
    """Embed each word in the pairs and returns the distance between thems"""
    sim_list = []
    for pair in pair_list:
        # calc cosine dist between w1 and w2
        emb1 = wemodel[pair[0]]
        emb2 = wemodel[pair[1]]
        similarity = np.dot(emb1, emb2) / (np.linalg.norm(emb1) * np.linalg.norm(emb2))
        sim_list.append(similarity)
        
    return sim_list


def get_pairs_dist_euc(pair_list, wemodel):
    """Embed each word in the pairs and returns the distance between thems"""
    dist_list = []
    for pair in pair_list:
        # calc cosine dist between w1 and w2
        emb1 = wemodel[pair[0]]
        emb2 = wemodel[pair[1]]
        dist = np.linalg.norm(emb1 - emb2)
        dist_list.append(dist)
        
    return dist_list


def get_word_list_for_clustering(word_dict):
    """Lemmatizing and remove stop words"""
    word_list = []
    for key, val in word_dict.items():
        if (not val['protected']):  # Not protected 
            if val['lemma']:
                word_list.append(val['lemma'])
            else:
                word_list.append(key)
            
    return list(set(word_list))  # Remove duplicates 


def embed_corpus(word_dict, stop_list, wemodel):
    """ Embeds the corpus using glove """

    word_list = get_word_list_for_clustering(word_dict)
    word_index = get_word_index_for_clustering(word_list, stop_list)

    # Iterate over your dictionary of words and embed them using GloVe
    embedded_dict = {}
    for word in word_index.keys():
        if word not in stop_list: # stopwords.words('english'):
            try:
                embedded_dict[word] = wemodel[word]
            except KeyError:
                # If the word is not in the GloVe vocabulary, assign a default embedding or skip it
                pass
    return embedded_dict


def run_clustering(word_dict, stop_list, wemodel, cosine = True, eps = None, n_jobs = -1):
    """ Runs clustering """
    # point to think - min_points in cluster to be defined according to k?
    # Get embedding
    embedded_dict = embed_corpus(word_dict, stop_list, wemodel)
    # Convert to numpy array
    embeddings = np.array(list(embedded_dict.values()))

    if not eps:
        # Based on embedding distance / 2, since DBSCAN uses the epsilon as radius and not diameter.
        if cosine:
            eps = define_eps_cos(wemodel=wemodel) / 2
        else:
            eps = define_eps_euc(wemodel=wemodel) / 2 
            # eps = find_eps_val(embeddings, cosine=cosine)  # Based on knee method

    logging.debug(f'Pre-defined epsilon for DBSCAN: {eps}.   Cosine metric: {cosine}.   n_jobs: {n_jobs}')
    
    # Chose 3 a min words per cluster (maybe reduce to 2?) Maybe according to k
    if cosine:
        dbscan = DBSCAN(eps=eps, min_samples=2, metric='cosine', n_jobs=n_jobs) 
    else:
        dbscan = DBSCAN(eps=eps, min_samples=2, n_jobs=n_jobs)  # Using Euclidian distance

    dbscan.fit(embeddings)

    labels = dbscan.labels_
    clusters = {}
    for i, key in enumerate(embedded_dict.keys()):
        cluster = labels[i]
        if cluster not in clusters:
            clusters[cluster] = []
        clusters[cluster].append(key)
    
    # Getting the maximal distance in each cluster
    distance_dict = get_dist_dict(embedded_dict, clusters, labels)

    return clusters, distance_dict, labels


def get_dist_dict(embedded_dict, clusters, labels):
    """Calculates the max distance fore each cluster and return a dicionary of cluster num: max. distance"""
    # init dict of dist:
    dist_dict = {}

    # for each cluster return the pair of words and max dist of the cluster:
    for ind in set(labels):
        filtered_dict = {k: v for k, v in embedded_dict.items() if k in clusters[ind]} # dict of embeddings of the words in the cluster
        curr_vecs = list(filtered_dict.values())

        # Finding the centorid
        centroid = np.mean(curr_vecs, axis=0)
        curr_dis_list = []
        for v in curr_vecs:
            # Adding the absolut distance from centorid
            curr_dis_list.append(np.abs(v-centroid)) 
        dist = np.mean(curr_dis_list)
        dist_dict[ind] = dist
    return dist_dict


def get_word_index_for_clustering(all_words, stop_list):
    """ Uses tokenizer to get word indexes """
    word_index = {}
    i = 0
    for word in all_words:
        if word and (word not in stop_list): #  stopwords.words('english')):
            word_index[word] = i
            i += 1

    return word_index


def plot_cluster_size_distribution(clusters):  # PIE CHART
    """
    Plots cluster size distribution.
    Input: dictionary of key: [item1, item2, ...]
    """
    # NOT INCLUSING CLUSTER -1
    size_list = []
    copy_clusters = clusters
    if -1 in copy_clusters:
        del copy_clusters[-1]

    for l in copy_clusters.values():
      #print(l)
      if l == -1:
        continue
      else:
        size_list.append(len(l))

    # cluster_sizes is a list or array containing the number of data points in each cluster
    plt.pie(size_list, labels=size_list)
    # plt.pie(size_list, labels=range(len(size_list)))

    plt.show()
    # cluster_sizes is a list or array containing the number of data points in each cluster
    # Set the maximum number of bins you want to display
    max_bins = 10

# Get the limited list of sizes based on the maximum number of bins
    limited_size_list = size_list[:max_bins]
    plt.bar(range(len(limited_size_list)), limited_size_list)
    # Customize the x-axis labels if needed
    plt.xticks(range(len(limited_size_list)), range(1, len(limited_size_list) + 1))
    plt.xlabel('Cluster No.')
    plt.ylabel('Words in cluster')
    plt.show()

if __name__ == '__main__':
    print('YEHH')

